/*
 * Copyright (C) 2002 Jan Panteltje <panteltje@yahoo.com>,
 * based on dvdauthor by Scott Smith (trckjunky@users.sourceforge.net) 
 * added multiple audio and subpicture channels.
 *
 * Modified and copy right Jan Panteltje 2002
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <getopt.h>

#include "dvdauthor.h"


static const char RCSID[] = "$Id: //depot/dvdauthor/src/ifogen.c#33 $";

#define SCRTIME 27000000

// must be a power of 2
// was 16
#define CBSIZE 16


struct circbuf
{
char buf[CBSIZE];
int ptr;
};


struct audpts
{
int pts;
int sect;
};

// was 100
#define MAXAUDPTS 100


struct vobuinfo
{
int sector,
	lastsector,
	scr,
	fsect,
	fnum,
	cellid,
	transpose;
int packts[2],
	pts[2],
	dts[2],
	numref,
	numaudpts;
int lastrefsect[3]; // why on earth do they want the LAST sector of the ref (I, P) frame?
struct audpts audpts[MAXAUDPTS]; // should be big enough...
char sectdata[2048]; // so we don't have to reread it
};


struct vob_summary
{
int numvobus;
long long lasttime;
struct vobuinfo *vi;
};


int audioflag = -1, audiowarn = 0; // 0 = ac3, 2 = mp1 layer 2, 4 = pcm
int tvwarn = 0;
int aspectflag = -1, aspectwarn = 0; // 0 = 4:3, 1 = 16:9
int resolutionflag = -1, resolutionwarn = 0; // 0 = 720xfull, 1 = 704xfull, 2 = 352xfull, 3 = 352xhalf

#define NUMCELLS(va) ((va)->vi[(va)->numvobus-1].cellid)


char *audio_language_str, *subpicture_language_str;


void set_ts(int *ts,int v)
{
if( ts[0] == -1 || v < ts[0] ) ts[0] = v;

if( ts[1] == -1 || v > ts[1] ) ts[1] = v;
} /* end function set_ts */


unsigned char cbyte(struct circbuf *c, int o)
{
return c->buf[(c->ptr + o + CBSIZE) & (CBSIZE - 1)];
} /* end function cbyte */
 

void cput(struct circbuf *c, int ch)
{
c->buf[c->ptr] = ch;
c->ptr = (c->ptr + 1) & (CBSIZE - 1);
} /* end function cput */


unsigned int buildtime(long long num, long long denom)
{
int rate = (tvflag == 1 ? 1 : 3);
double frates[4] = {0, 25, 0, 29.97};
int hr, min, sec, fr;

sec = num / denom;
min = sec / 60;
hr = min / 60;
min %= 60;
sec %= 60;
hr = (hr / 10) * 16 + hr % 10;
min = (min / 10) * 16 + min % 10;
sec = (sec / 10) * 16 + sec % 10;
num %= denom;
fr = ((int)(num * frates[rate] / denom) );
fr = (fr / 10) * 16 + fr % 10;
return (hr << 24) | (min << 16) | (sec << 8) | fr | (rate << 6);
} /* end function buildtime */


void write4(char *p,unsigned int v)
{
p[0] = (v >> 24) & 255;
p[1] = (v >> 16) & 255;
p[2] = (v >> 8) & 255;
p[3] = v & 255;
} /* end function write4 */


void write2(char *p, unsigned int v)
{
p[0] = (v >> 8) & 255;
p[1] = v & 255;
} /* end function write2 */


unsigned int read2(unsigned char *p)
{
return (p[0] << 8) | p[1];
} /* end function read2 */


void updatevideoattr(char *name, int *v, int nv, int *w)
{
if( nv == -1 ) return;

if(v[0] != -1 && v[0] != nv && w[0] == 0 )
	{
	fprintf(stderr,\
	"WARN: attempt to update %s from %d to %d; skipping\n",\
	name, v[0], nv);
	w[0] = 1;
    }

if(v[0] == -1 ) v[0] = nv;
} /* end function updatevideoattr */


void updateframerate(int frame)
{
int pal;

switch(frame)
	{
    case 3: //   "25.0 (PAL / SECAM VIDEO / converted FILM)"
        pal = 1;
        break;

    case 4: //   "30000.0 / 1001.0 (NTSC VIDEO)"
        pal = 0;
        break;
    
    default:
        pal = -1;
        break;
    }

updatevideoattr("pal flag", &tvflag, pal, &tvwarn);
} /* end function updateframerate */


void updateresolution(int h, int v)
{
int res;

if( h >= 720 ) res = 0;
else if ( h >= 704 ) res = 1;
else if ( v >= 384 ) res = 2;
else res = 3;

updatevideoattr("resolution flag", &resolutionflag, res, &resolutionwarn);
} /* end function updateresolution */


void setattr()
{
set_video_mode();

if( tvflag == -1 ) tvflag = 0;
if( aspectflag == -1 ) aspectflag = 0;
if( resolutionflag == -1 ) resolutionflag = 0;
if( audioflag == -1 ) audioflag = 2;

videoattr |= (tvflag * 0x1000) | (aspectflag * 0xc00) | (resolutionflag << 2);

fprintf(stderr,"INFO: Generating VTS with the following video attributes:\n");
fprintf(stderr,"INFO: TV standard: %s\n", tvflag ? "PAL" : "NTSC");
fprintf(stderr,"INFO: Aspect ratio: %s\n", aspectflag ? "16:9" : "4:3");
fprintf(stderr,"INFO: Resolution: %dx%d\n",
	resolutionflag ? (resolutionflag == 1 ? 704 : 352) : 720,
	(resolutionflag == 3 ? 240 : 480) * (tvflag ? 6 : 5) / 5);

fprintf(stderr,\
"INFO: Audio format: %s\n", audioflag == 0 ? "ac3" : (audioflag == 2 ? "mp2" : (audioflag == 4 ? "pcm" : "-")));
} /* end function setattr */


int timeline[19] = {1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 20, 60, 120, 240};

int findvobu(struct vob_summary *va, int pts, int roundup)
{
int i;

for( i = 0; i < va->numvobus; i++ )
	{
	if( va->vi[i].pts[0] == pts ) return i;
	if( pts < va->vi[i].pts[0] ) return i - 1;
	if( pts < va->vi[i].pts[1] ) return i + roundup;
    }

return -1;
} /* end function findvobu */


unsigned int getsect(struct vob_summary *va,int vobnum,int skip,int sectoff,unsigned notfound)
{
int dessect;

if( skip ) skip=0x40000000;
if( vobnum < 0 || vobnum >= va->numvobus ) return notfound | skip;

dessect = va->vi[vobnum].sector;
if( dessect < sectoff ) return (sectoff - dessect) | 0x80000000 | skip;
else return (dessect - sectoff) | 0x80000000 | skip;
} /* end function getsect */


int readscr(const unsigned char *buf)
{
return ((buf[0] & 0x18) << 27)|
	((buf[0] & 3) << 28) |
	(buf[1] << 20) |
	((buf[2] & 0xf8) << 12) |
	((buf[2] & 3) << 13) |
	(buf[3] << 5) |
	((buf[4] & 0xf8) >> 3);
} /* end function readscr */


void writescr(unsigned char *buf,unsigned int scr)
{
buf[0] = ((scr >> 27) & 0x18) | ((scr >> 28) & 3) | 68;
buf[1] = scr >> 20;
buf[2] = ((scr >> 12) & 0xf8) | ( (scr >> 13) & 3) | 4;
buf[3] = scr >> 5;
buf[4] = ((scr << 3) & 0xf8) | (buf[4] & 7);
} /* end function writescr */


int readpts(const unsigned char *buf)
{
int a1, a2, a3, pts;

a1 = (buf[0] & 0xe) >> 1;
a2 = ( (buf[1] << 8) | buf[2]) >> 1;
a3 = ( (buf[3] << 8) | buf[4]) >> 1;

pts = ( ( (long long)a1) << 30) | (a2 << 15) | a3;
return pts;
} /* end function readpts */


void writepts(unsigned char *buf,unsigned int pts)
{
// this preserves the PTS / DTS / PTSwDTS top bits
buf[0] = ( (pts >> 29) & 0xe) | (buf[0] &0xf1);
write2(buf + 1, (pts >> 14) | 1);
write2(buf + 3, (pts << 1) | 1);
} /* end function writepts */


void FixVobus(char *fbase, struct vob_summary *va)
{
int h = -1;
int i, j, scr, fnum = -1;
int vff, vrew;
char *buf = 0;

for( i = 0; i < va->numvobus; i++ )
	{
	if( va->vi[i].fnum != fnum )
		{
		char fname[200];
		if( h >= 0 ) close(h);
		fnum = va->vi[i].fnum;
		sprintf(fname, "%s_%d.VOB", fbase,fnum);
		h=open(fname, O_WRONLY);
		if( h < 0 )
			{
			fprintf(stderr,\
			"ERR:  Error opening %s: %s\n", fname, strerror(errno));
			exit(1);
			}
		}
	buf = va->vi[i].sectdata;

	scr = readscr(buf + 4);

	buf[0x2c] = 0;
	write4(buf + 0x2d, va->vi[i].sector);

	/* 0x31 (2 bytes) vobu_cat, flags, including APS analog protection system */
	buf[0x31] = 0;
	buf[0x32] = 0;

	/* 0x33 (2 bytes) reserved */
	buf[0x33] = 0;
	buf[0x34] = 0;

	/* 
	0x35 (4 bytes, 20 flags) vobu_uop_ctl, 
	bitmask prohibited user operations, flag 20 is audio channel change inhibit.
	*/
	buf[0x35] = 0;
	buf[0x36] = 0;
	buf[0x37] = 0;
	buf[0x38] = 0;

	write4(buf + 0x39, va->vi[i].pts[0]); // vobu_s_ptm
	write4(buf + 0x3d, va->vi[i].pts[1]); // vobu_e_ptm


//va->vi[i].transpose=120 * 90000; // new OK
/*
	j = 0;
	while(j < va->numvobus && va->vi[j].cellid != i + 1) j++;
	buf[16 + read2(buf + 248) + i * 24] = 2 | (i ? 8 : 0);
	firsttime = va->vi[j].packts[0];
*/

//fprintf(stdout, "WAS -1 va->vi[0].packts[0]=%.2f\n", (double)va->vi[0].packts[0] / 90000.0);
//	write4(buf + 0x45, buildtime(scr - va->vi[i].transpose, 90000)); // total guess
	write4(buf + 0x45, buildtime(scr - va->vi[0].packts[0], 90000)); // new

	buf[0x406] = 1;
	write4(buf + 0x407, scr);
	write4(buf + 0x40b, va->vi[i].sector); // current lbn
	for( j = 0; j < va->vi[i].numref; j++ )
		write4(buf + 0x413 + j * 4, va->vi[i].lastrefsect[j] - va->vi[i].sector);
	write2(buf + 0x41f, va->vi[i].cellid);

	buf[0x422] = 1;
	write4(buf + 0x423, buildtime(scr, 90000)); // total guess

	write4(buf + 0x433, va->vi[0].pts[0]);
	write4(buf + 0x437, va->vi[va->numvobus - 1].pts[1]);
	write4(buf + 0x4f1, getsect(va, i + 1, 0, va->vi[i].sector, 0xbfffffff));
	write4(buf + 0x541, getsect(va, i + 1, 0, va->vi[i].sector, 0x3fffffff));
	write4(buf + 0x545, getsect(va, i - 1, 0, va->vi[i].sector, 0x3fffffff));
	write4(buf + 0x595, getsect(va, i - 1, 0, va->vi[i].sector, 0xbfffffff));

	for( j = 0; j < va->vi[i].numaudpts; j++ )
	// audio chunks for mp2 are 1/75th of a second, so add that much padding
		if( va->vi[i].audpts[j].pts + 90000 / 75 >= va->vi[i].pts[0] )
			break;

	if( j > 0 && j < va->vi[i].numaudpts )
		write2(buf + 0x599, va->vi[i].audpts[j - 1].sect - va->vi[i].sector);

	write4(buf + 0x40f, va->vi[i].lastsector - va->vi[i].sector);

	vff = i;
	vrew = i;
	for( j = 0; j < 19; j++ )
		{
		int nff, nrew;

		nff = findvobu(va, va->vi[i].pts[0] + timeline[j] * 45045, 1);
		nrew = findvobu(va, va->vi[i].pts[0] - timeline[j] * 45045 ,0);
		write4(buf + 0x53d - j * 4, getsect(va, nff, nff > vff + 1, va->vi[i].sector, 0x3fffffff));
		write4(buf + 0x549 + j * 4, getsect(va, nrew, nrew < vrew - 1, va->vi[i].sector, 0x3fffffff));
		vff = nff;
		vrew = nrew;
		}

    lseek(h, va->vi[i].fsect * 2048, SEEK_SET);

	write(h, buf, 2048);

	if( !(i & 15) ) 	
		fprintf(stderr,\
		"STAT: fixing VOBU at %dMB (%d/%d, %d%%)\r",\
		va->vi[i].sector / 512, i, va->numvobus, i * 100 / va->numvobus);
	} /* end for i */

close(h);
fprintf(stderr,"\n");
} /* end function FixVobus */


void transpose_ts(unsigned char *buf, int tsoffs)
{
// pack scr
if( buf[0] == 0 && buf[1] == 0 && buf[2] == 1 && buf[3] == 0xba )
	{
	writescr(buf + 4, readscr(buf + 4) + tsoffs);

	// video/audio?
	// pts?
	if( buf[14] == 0 && buf[15] == 0 && buf[16] == 1 &&
	(buf[17] >= 0xc0 && buf[17] < 0xf0) &&
	(buf[21] & 128) )
		{
		writepts(buf + 23, readpts(buf + 23) + tsoffs);

		// dts?
		if( buf[21] & 64 )
			{
			writepts(buf + 28, readpts(buf + 28) + tsoffs);
			}
		}
	}
} /* end function transpose_ts */


static unsigned char bigwritebuf[16*2048];
static int writebufpos=0;
static int writefile=-1;

void writeflush()
{
if( !writebufpos ) return;

if( write(writefile,bigwritebuf,writebufpos) != writebufpos )
	{
	fprintf(stderr,"ERR:  Error writing data\n");
	exit(1);
	}

writebufpos=0;
} /* end function writeflush */


unsigned char *writegrabbuf()
{
unsigned char *buf;

if( writebufpos == sizeof(bigwritebuf) ) writeflush();

buf = bigwritebuf + writebufpos;
writebufpos += 2048;

return buf;
} /* end function writegrabbuf */


void writeundo()
{
writebufpos -= 2048;
} /* end function writeundo */


void writeclose()
{
writeflush();

if( writefile != -1 )
	{
	close(writefile);
	writefile = -1;
	}
} /* end function writeclose */


void writeopen(const char *newname)
{
writefile = open(newname, O_CREAT | O_WRONLY, 0666);

if( writefile < 0 ) 
	{
	fprintf(stderr,\
	"ERR:  Error opening %s: %s\n", newname, strerror(errno) );
	exit(1);
	}
} /* end function writeopen */



int FindVobus(char *fbase, int numfiles, char **fname, struct vob_summary *va)
{
unsigned char *buf;
FILE *h;
int cursect=0, fsect = -1, fnum, outnum = 0;
int lastrefsect = -1, tsoffs = 0;
int backfillpts = 0;
struct circbuf cb;
int i, j;

va->numvobus = 0;
va->vi = 0;
va->lasttime = 0;
cb.ptr = 0;
for(fnum = 0; fnum < numfiles; fnum++)
	{
	if( va->numvobus )
		{
		tsoffs = va->vi[va->numvobus - 1].packts[1] + 90000;
        }

//tsoffs = 90000 * 120;
//fprintf(stderr, "tsoffs=%.2f\n", (double)tsoffs / 90000.0);

	fprintf(stderr, "\nSTAT: Processing %s...\n", fname[fnum]);
	h = fopen(fname[fnum], "r");
	if( !h )
		{
		fprintf(stderr,\
		"ERR:  Error opening %s: %s\n", fname[fnum], strerror(errno));
		exit(1);
		}
	while(1)
		{
        if( fsect == 524272 )
			{
            writeclose();
            fsect = -1;
            }

		buf = writegrabbuf();
		if( 2048 != fread(buf, 1, 2048, h) )
			{
			writeundo();
			break;
            }

		transpose_ts(buf, tsoffs);

		if( fsect == -1 )
			{
			char newname[200];
			outnum++;
			fsect = 0;
			sprintf(newname, "%s_%d.VOB", fbase, outnum);
			writeopen(newname);
			}

		/* test for system header */
		if( buf[14] == 0 &&
		buf[15] == 0 &&
		buf[16] == 1 &&
		buf[17] == 0xbb && // system header

		// no private stream 1 in mplex or tcmplex ??
		buf[38] == 0 &&
		buf[39] == 0 &&
		buf[40] == 1 &&
		buf[41] == 0xbf && // 1st private2

		buf[1024] == 0 &&
		buf[1025] == 0 &&
		buf[1026] == 1 &&
		buf[1027] == 0xbf // 2nd private2
		)
			{
           	va->numvobus++;
			va->vi = \
				realloc(va->vi, va->numvobus * sizeof(struct vobuinfo) );
			va->vi[va->numvobus - 1].sector = cursect;
			va->vi[va->numvobus - 1].fsect = fsect;
			va->vi[va->numvobus - 1].fnum = outnum;
			va->vi[va->numvobus - 1].cellid = fnum + 1;
			va->vi[va->numvobus - 1].scr = readscr(buf + 4);
			va->vi[va->numvobus - 1].transpose = tsoffs;
			va->vi[va->numvobus - 1].pts[0] = -1;
			va->vi[va->numvobus - 1].pts[1] = -1;
			va->vi[va->numvobus - 1].packts[0] = -1;
			va->vi[va->numvobus - 1].packts[1] = -1;
			va->vi[va->numvobus - 1].numref = 0;
			va->vi[va->numvobus - 1].numaudpts = 0;

			memcpy(va->vi[va->numvobus - 1].sectdata, buf, 2048);

			if( !(va->numvobus & 15) )
				fprintf(stderr,"STAT: VOBU %d at %dMB\r",\
				va->numvobus, cursect / 512);

			lastrefsect = -1;
			backfillpts = 1;
			} /* end if system header */

		if(va->numvobus == 0)
			{
			fprintf(stderr, "incorrect header\n");

			for(i = 0; i < 2048; i += 8)
				{
				for(j = 0; j < 8; j++)
					{
					fprintf(stderr, "b[%d]=%02x ", i + j, buf[i + j]); 
					} /* end for j */

				fprintf(stderr, "\n");
				} /* end for i */

			exit(1);
			}

		va->vi[va->numvobus - 1].lastsector = cursect;

		/* test for video stream */

		if( buf[0] == 0 &&
		buf[1] == 0 &&
		buf[2] == 1 &&
		buf[3] == 0xba &&
		buf[14] == 0 &&
		buf[15] == 0 &&
		buf[16] == 1 &&
		buf[17] == 0xe0 )
			{
			// video
			int i;
			for( i = 0x17 + buf[0x16];\
			i < 0x14 + buf[0x12] * 256 + buf[0x13]; i++ )
				{
				cput(&cb, buf[i]);
				if( cbyte(&cb, -6) == 0 &&
				cbyte(&cb, -5) == 0 &&
				cbyte(&cb, -4) == 1 &&
				cbyte(&cb, -3) == 0 )
					{
					int ptype = (cbyte(&cb, -1) >> 3) & 7;
					if( lastrefsect != -1 && va->vi[va->numvobus - 1].numref < 3 )
						{
						va->vi[va->numvobus - 1].lastrefsect[va->vi[va->numvobus - 1].numref++] = cursect;
						lastrefsect = -1;
						}
					if( ptype == 1 || ptype == 2 ) // I or P frame
						lastrefsect = cursect;
					}

				if( cbyte(&cb, -8) == 0 &&
				cbyte(&cb, -7) == 0 &&
				cbyte(&cb, -6) == 1 &&
				cbyte(&cb, -5) == 0xb3 )
					{
					int hsize, vsize, aspect, frame;
                        
					hsize = (cbyte(&cb, -4) << 4) | (cbyte(&cb, -3) >> 4);
					vsize = ((cbyte(&cb,-3) << 8) & 0xf00) | cbyte(&cb, -2);
					aspect = cbyte(&cb,-1) >> 4;
					frame = cbyte(&cb, -1) & 0xf;

					if( aspect == 2 ) aspect = 0;
					else if( aspect == 3 ) aspect = 1;
					else aspect = -1;

					updateframerate(frame);
					updateresolution(hsize, vsize);
					updatevideoattr("aspect ratio", &aspectflag, aspect, &aspectwarn);
					}
				}
			if( buf[21] & 128 )
				{
				// check whether there's a pts
				struct vobuinfo *vi = &va->vi[va->numvobus - 1];

				int pts = readpts(buf + 23);
// fprintf(stderr,"pts %d (%02x %02x %02x %02x %02x)\n",pts,buf[23],buf[24],buf[25],buf[26],buf[27],buf[28]);

				set_ts(vi->pts, pts);
				if( backfillpts && va->numvobus >= 2 )
					{
					set_ts(va->vi[va->numvobus - 2].pts, pts);
					}
				}
			backfillpts = 0;
			} /* end if video stream */

		/* test for audio stream */

		if( buf[0] == 0 &&
		buf[1] == 0 &&
		buf[2] == 1 &&
		buf[3] == 0xba &&
		buf[14] == 0 &&
		buf[15] == 0 &&
		buf[16] == 1 &&
//		buf[17] == 0xc0 &&
		(buf[17] >= 0xc0 && buf[17] < 0xf0) &&
		buf[21] & 128 )
			{
			struct vobuinfo *vi = &va->vi[va->numvobus - 1];
			int pts = readpts(buf + 23);

			if( vi->numaudpts < MAXAUDPTS ) 
				{
				vi->audpts[vi->numaudpts].pts = pts;
				vi->audpts[vi->numaudpts].sect = cursect;
				vi->numaudpts++;
				if( vi->numaudpts >= MAXAUDPTS )
					{
					fprintf(stderr,\
					"WARN: Reached %d audio pts for this VOBU\n", MAXAUDPTS);
					}
				}
			} /* end if audio stream */

		// read time from pack header
		if( buf[0] == 0 && buf[1] == 0 && buf[2] == 1 && buf[3] == 0xba )
			{
			long scr, scrhi, scrext;
			long long fulltime;

			scrhi = (buf[4] & 0x20) >> 5;
			scr = ( (buf[4] & 0x18) << 27) | ( (buf[4] & 3) << 28) |
			(buf[5] << 20) | ( (buf[6] & 0xf8) << 12) |
			( (buf[6] & 3) << 13) | (buf[7] << 5) | ( (buf[8] & 0xf8) >> 3);

			scrext = ( (buf[8] & 3) << 7) | (buf[9] >> 1);

			fulltime = ( (long long)scrhi) << 32 | ( (long long)scr);
			fulltime *= 300;
			fulltime += scrext;

			if( fulltime > va->lasttime ) va->lasttime = fulltime;
			set_ts(va->vi[va->numvobus - 1].packts, fulltime / 300);
        	}
		cursect++;
		fsect++;
		} /* end while */
	fclose(h);
	} /* end for numfiles */

writeclose();

fprintf(stderr,\
"STAT: VOBUS: %d, %dMB            \n",\
va->numvobus, cursect / 512);

fprintf(stderr,\
"STAT: last time=%lld.%02lld sec\n",\
va->lasttime / SCRTIME, (va->lasttime % SCRTIME) / (SCRTIME / 100));

return 1;
} /* end function FindVobus */


int numsectTMAPT(struct vob_summary *va)
{
int ptsspan = va->vi[va->numvobus-1].pts[1] - va->vi[0].pts[0];
int sz = 16 + 4 * (ptsspan / 90090 - 1);

return (sz + 2047) / 2048;
} /* end function numsectTMAPT */


void CreateTMAPT(FILE *h, struct vob_summary *va)
{
int ptsspan = va->vi[va->numvobus - 1].pts[1] - va->vi[0].pts[0];
int numtmapt, i;
char buf[16];

numtmapt = ptsspan / 90090 - 1;
memset(buf, 0, 16);
buf[1] = 1;
write4(buf + 4, 15 + numtmapt * 4);
buf[11] = 0xc;
buf[12] = 1;
write2(buf + 14, numtmapt);
fwrite(buf, 1, 16, h);

for( i = 0; i < numtmapt; i++ )
	{
	int vobu = findvobu(va, va->vi[0].pts[0] + (i + 1) * 90090, 1);
	write4(buf, va->vi[vobu].sector);
	fwrite(buf, 1, 4, h);
	}

i = (16 + 4 * numtmapt) & 2047;
if( i )
	{
	i = 2048 - i;
	memset(buf, 0, 16);
	while(i >= 16)
		{
		fwrite(buf, 1, 16, h);
		i -= 16;
		}
	if( i ) fwrite(buf, 1, i, h);
	}
} /* end function CreateTMAPT */
 

int numsectVOBUAD(struct vob_summary *va)
{
int sz = 4 + va->numvobus * 4;

return (sz + 2047) / 2048;
} /* end function numsectVOBUAD */


void CreateVOBUAD(FILE *h,struct vob_summary *va)
{
int i;
char buf[16];

write4(buf, va->numvobus * 4 + 3);
fwrite(buf, 1, 4, h);

for( i = 0; i < va->numvobus; i++ )
	{
	write4(buf, va->vi[i].sector);
	fwrite(buf, 1, 4, h);
	}

i = (4 + va->numvobus *4 ) & 2047;
if( i )
	{
	i = 2048 - i;
	memset(buf, 0, 16);
	while(i >= 16)
		{
		fwrite(buf, 1, 16, h);
		i -= 16;
		}
	if( i ) fwrite(buf, 1, i, h);
	}
} /* end function CreateVOBUAD */


void WriteIFO(FILE *h, struct vob_summary *va, int bup_mode)
{
int a;
static char buf[2048];
FILE *fptr;
char temp[4096];

#ifdef ORIG_COLORS
static int colors[16] =\
{
0x00108080,
0x00b48080,
0x007d8080,
0x00eb8080,
0x0051f05a,
0x00902235,
0x00286df0,
0x00ca4892,
0x00aaaea4,
0x00de8948,
0x00aeb76d,
0x0049a6dc,
0x00616389,
0x0098bc51,
0x0037b389,
0x00718947
};
#endif

static int colors[16] =\
{
0x00008080,
0x00f08080,
0x00008080,
0x00008080,
0x0051f05a,
0x00902235,
0x00286df0,
0x00ca4892,
0x00aaaea4,
0x00de8948,
0x00aeb76d,
0x0049a6dc,
0x00616389,
0x0098bc51,
0x0037b389,
0x00718947
};

int i;

// sect 0: VTS toplevel
memset(buf, 0, 2048);

/* 0x00 (12 bytes) id string */
memcpy(buf, "DVDVIDEO-VTS", 12);

/* 0x0c (4 bytes) last sector of title set (last sector of BUP) */
write4(buf + 12, 8 + 2 * (numsectTMAPT(va) + numsectVOBUAD(va)) + \
va->vi[va->numvobus - 1].lastsector); // total guess...

/* 0x1c (4 bytes) last sector of IFO */
buf[31] = 3 + numsectTMAPT(va) + numsectVOBUAD(va);

/* 0x20 (2 bytes) version number (byte0 reserved, byte 1, major, minor) */
/* version 1.1 ?? */
buf[33] = 0x11;

/* 0x40 in VMG IFO provider ID 32 bytes */

/* 0x80 (4 bytes) end byte address of VTS_MAT */
write4(buf + 128, 0x7ff);

/* 0xc4 start sector of Title Vob */
buf[199] = 4 + numsectTMAPT(va) + numsectVOBUAD(va); //0xc7

/* 0xc8 sector pointer to VTS_PTT_SRPT (table of Titles and Chapters) */
buf[203] = 1;						//0xcb

/* 0xcc sector pointer to VTS_PGCI (Title Program Chain table) */
buf[207] = 2;						//0xcf

/* 0xd4 sector pointer to VTS_TMAPTI (time map) */
buf[215] = 3;						//0xd7

/* 0xe0 sector pointer to VTS_C_ADT (title set cell address table) */
buf[227] = 3 + numsectTMAPT(va);	//0xe3

/* 0xe4 sector pointer to VTS_VOBU_ADMAP (title set VOBU address map) */
buf[231] = 4 + numsectTMAPT(va);	//0xe7

/* video attributes to this buffer buf[] */
//	write2(buf + 512, videoattr);

set_video_mode();
buf[256] = video_mode0;	//0x200
buf[257] = video_mode1;	//0x201

buf[512] = video_mode0;	//0x200
buf[513] = video_mode1;	//0x201

set_audio_mode(buf, 0, audio_language_str);
set_audio_mode(buf, 1, audio_language_str);

/* 580 0x244 next free */
set_subpicture_mode(buf, 0, subpicture_language_str);
set_subpicture_mode(buf, 1, subpicture_language_str);

/* WRITE SECTOR 0 VTS toplevel */
fwrite(buf, 1, 2048, h);	


// sect 1: VTS_PTT_SRPT (table of titles and chapters)
memset(buf, 0, 2048);
buf[1] = 1; // # of titles
buf[11] = 0xc;

for(i = 0; i < NUMCELLS(va); i++ )
	{
	buf[13 + i * 4] = 1;
	buf[15 + i * 4] = i + 1;
	}

buf[7] = 11 + i * 4;

/* WRITE SECTOR 1 VTS_PTT_SRPT */
fwrite(buf, 1, 2048, h);
   

// sect 2: VTS_PGCI (Title Program Chain table)
memset(buf, 0, 2048);

/* number of program chains */
buf[1] = 1;

buf[0x06] = 1;
buf[0x07] = 11 + 28 * NUMCELLS(va) + ( (NUMCELLS(va) + 1) & 0xFE);

/* 
high nibble offset to VTS_PGC, relative to VTS_PGCI (here 8)
low nibble ?? 
*/
buf[0x08] = 0x81;


/* ?? */
buf[0x0f] = 0x10;

/* PGC header */
/* offset 2 (1 byte) number of programs */
buf[0x12] = NUMCELLS(va);

/* offset 3 (1 byte) number of cells*/
buf[0x13] = NUMCELLS(va);

/*
offset 4 (4 bytes)
playback time BCD, hh:mm:ss:ff with bits 7&6 of frame (last) byte indicating frame rate
11 = 30 fps, 10 = illegal, 01 = 25 fps, 00 = illegal
*/
write4(buf + 0x14, buildtime(va->lasttime, SCRTIME));

/* offset 8 (4 bytes) prohibited user ops */
buf[0x18] = 0;
buf[0x19] = 0;
buf[0x1a] = 0;
buf[0x1b] = 0;

if(debug_flag)
	{
	fprintf(stderr, "ifogen: audio_streams=%d\n", audio_streams);
	}

/* 
NOTE: 
modes (set audio) is called BEFORE this and sets audio_streams.
This enables (bit 7) the selection of each audio channel.
*/

/* offset 0C true 0x1c PGC_AST_CTL 8 * 2 Audio Stream Control */

/* ch 1 offset 0x0c 0x0d */
buf[0x1c] = 0;
if(audio_streams > 0) buf[0x1c] |= 0x80;

/* ch 2 offset 0x0e 0x0f */
buf[0x1e] = 1;
if(audio_streams > 1) buf[0x1e] |= 0x80; 

/* ch 3 offset 0x10 0x11 */
buf[0x20] = 2;
if(audio_streams > 2) buf[0x20] |= 0x80;

/* ch 4 offset 0x12 0x13 */
buf[0x22] = 3;
if(audio_streams > 3) buf[0x22] |= 0x80;

/* ch 5 offset 0x14 0x15 */
 buf[0x24] = 4;
if(audio_streams > 4) buf[0x24] |= 0x80;

/* ch 6 offset 0x16 0x17 */
buf[0x26] = 5;
if(audio_streams > 5) buf[0x26] |= 0x80;

/* ch 7 offset 0x18 0x19 */
buf[0x28] = 6;
if(audio_streams > 6) buf[0x28] |= 0x80;

/* ch 8 offset 0x1a 0x1b */	
 buf[0x2a] = 7;
if(audio_streams > 7) buf[0x2a] |= 0x80;

/* end audio stream control block at offset 0x10 */


/* SUBTITLES info in pgc.html, PGC_SPST_CTL 32 * 4 bytes */
/* offset 0x1c true 0x2c (32 * 4 bytes) PGC_SPST_CTL subpicture control */
/* Subpicture Stream Control, each entry:
  byte  7                     6 5       4 3 2 1 0
  0     1 = stream available  reserved  Stream number for 4:3
  1     reserved(7-5)                   Stream number for wide
  2     reserved(7-5)                   Stream number for letterbox
  3     reserved(7-5)                   Stream number for pan&scan
*/

if(debug_flag)
	{
	fprintf(stderr, "ifogen, subpicture_streams=%d\n", subpicture_streams);
	}

for(i = 0; i < 32; i++)
	{
	buf[0x2c + (4 * i)] = i;

	if(subpicture_streams > i)
		{
		buf[0x2c + (4 * i)] |= 0x80;
		}			
	} /* end for i */
/* end subpictures */


/* if any subtitles load palette created by xste, if not found use internal table */

if(subpicture_streams)
	{
	sprintf(temp, "xste-palette.dat");
	fptr = fopen(temp, "r");
	if(! fptr)
		{
		fprintf(stderr,\
		"Could not open file %s for read, using internal palette (try running xste in DVD mode?).\n",
		temp);
	
		/* offset A4 true 0xB4 16 * 4 palette (0, Y, Cr, Cb) */
		for(i = 0; i < 16; i++ ) write4(buf + 180 + i * 4, colors[i]);
		}
	else
		{
		fprintf(stderr, "using palette from file %s.\n", temp);
		/* write out colors, the hex is the 0yuv combined in one integer 00yyuuvv */
		for(i = 0; i < 16; i++)
			{
			fscanf(fptr, "%x", &a);	

			write4(buf + 180 + i * 4, a);
			} /* end for i */
		fclose(fptr);
		}
	} /* end if subpicture_streams */
else /* no subtitles */
	{
	/* offset A4 true 0xB4 16 * 4 palette (0, Y, Cr, Cb) */
	for(i = 0; i < 16; i++ ) write4(buf + 180 + i * 4, colors[i]);
	}

buf[245] = 0xec;
write2(buf + 246, read2(buf + 244) + 0x10);
write2(buf + 248, read2(buf + 246) + ( (NUMCELLS(va) + 1) & 0xFE));
write2(buf + 250, read2(buf + 248) + NUMCELLS(va) * 24);
buf[255] = 1; // ???

/*
fprintf(stderr,\
"read2(buf + 244)=%d (0x%x) read2(buf + 246)=%d (0x%x) read2(buf + 248)=%d (0x%x\n",\
read2(buf + 244), read2(buf + 244), read2(buf + 246), read2(buf + 246), read2(buf + 248), read2(buf + 248) );
*/

for(i = 0; i < NUMCELLS(va); i++ ) buf[16 + read2(buf + 246) + i] = i + 1;

for(i = 0; i < NUMCELLS(va); i++ )
	{
	int j, firsttime;

	j = 0;
	while(j < va->numvobus && va->vi[j].cellid != i + 1) j++;

	buf[16 + read2(buf + 248) + i * 24] = 2 | (i ? 8 : 0);

	firsttime = va->vi[j].packts[0];
//fprintf(stderr, "WAS 1 i=%d j=%d firsttime=%.2f\n", i, j, (double)firsttime / 90000.0); // new

	write4(buf + 24 + read2(buf + 248) + i * 24, va->vi[j].sector);

	while(j + 1 < va->numvobus && va->vi[j + 1].cellid == i + 1) j++;

	write4(buf + 20 + read2(buf + 248) + i * 24, buildtime(va->vi[j].packts[1] - firsttime, 90000));
	write4(buf + 32 + read2(buf + 248) + i * 24, va->vi[j].sector);
	write4(buf + 36 + read2(buf + 248) + i * 24, va->vi[j].lastsector);
	}

for(i = 0; i < NUMCELLS(va); i++ )
	{
	buf[17 + read2(buf + 250) + i * 4] = i + 1;
	buf[19 + read2(buf + 250) + i * 4] = 1;
	}

/* WRITE SECTOR 2  VTS_PGCIT */
fwrite(buf, 1, 2048, h);


// sect 3: ??? VTS_TMAPT (time map)
// (and at offset ? title set cell address table)
CreateTMAPT(h,va);

// sect 4: Call address table
// at offset (title set VOBU address map)
memset(buf, 0, 2048);
buf[1] = NUMCELLS(va);
buf[7] = 7 + NUMCELLS(va) * 12;
for(i = 0; i < NUMCELLS(va); i++ )
	{
	int j;

	j = 0;
	while(j<va->numvobus && va->vi[j].cellid != i + 1) j++;

	buf[9 + i * 12] = i + 1;
	buf[10 + i * 12] = 1;
	write4(buf + 12 + i * 12, va->vi[j].sector);
	while(j + 1 < va->numvobus && va->vi[j + 1].cellid == i + 1) j++;
	write4(buf + 16 + i * 12, va->vi[j].lastsector);
	}

/* WRITE SECTOR 4 Call address table */
fwrite(buf,1 , 2048, h);

// sect 5: VOBU address map
CreateVOBUAD(h, va);
} /* end function WriteIFO */



void WriteIFOs(char *fbase,struct vob_summary *va)
{
FILE *h;
static char buf[1000];

sprintf(buf, "%s_0.IFO", fbase);
h=fopen(buf, "w");
WriteIFO(h, va, 0);
fclose(h);

sprintf(buf, "%s_0.BUP", fbase);
h = fopen(buf,"w");
WriteIFO(h, va, 1);
fclose(h);
} /* end function WriteIFOs */


int main(int argc,char **argv)
{
struct vob_summary va;
char *fbase = 0;
static char realfbase[1000];
int vtsnum = 1;

static struct option longopts[] = {
	{"audio_language_codes", 1, 0, 'l'},
	{"resolution", 1, 0, 'r'},
	{"subpicture_languages_codes", 1, 0, 's'},
	{"tv", 1, 0, 't'},
	{"aspect-ratio", 1, 0, 'a'},
	{"audio-format", 1, 0, 'A'},
	{"vts", 1, 0, 'v'},
	{"next-vts", 0, 0, 'n'},
	{0, 0, 0, 0}
	};


/* defaults */
tvflag = -1;

while(1)
	{
	int c = getopt_long(argc, argv, "o:l:s:", longopts, NULL);
	if( c == -1 )
	break;

	switch(c)
		{
		case 'l':
		audio_language_str = strsave(optarg);
		if(! audio_language_str)
			{
			exit(1);
			}		
			break;
		case 'o': fbase = optarg;
			break;
		case 'r':
			{
			int h, v;

			if( 2 != sscanf(optarg,"%dx%d", &h, &v) )
				{
				fprintf(stderr, "ERR:  Unable to parse resolution: %s\n", optarg);
				return 1;
				}

			if( resolutionflag != -1 )
				{
				fprintf(stderr, "ERR:  Already defined resolution\n");
				return 1;
				}
			updateresolution(h, v);
			break;
			}
		case 's':
			if(strcmp(optarg, "none") == 0)
				{
				fprintf(stderr, "No subpicture streams.\n");
				subpicture_language_str = 0;
				}
			else
				{
				subpicture_language_str = strsave(optarg);
				if(! subpicture_language_str)
					{
					exit(1);
					}		
				}
//fprintf(stderr, "IFOGEN subpicture_language_str=%s\n", subpicture_language_str);
			break;
		case 't':
			{
			int p = -1;
			if( !strcasecmp(optarg, "pal") )
			p = 1;
			else if (!strcasecmp(optarg, "ntsc") )
			p = 0;
			else
				{
				fprintf(stderr, "ERR:  Unrecognized TV standard: %s\n", optarg);
				return 1;
				}

			if( tvflag != -1 )
				{
				fprintf(stderr, "ERR:  Already defined TV standard.\n");
				return 1;
				}

			tvflag = p;
			break;
			}
		case 'a':
			{
			int a = -1;
			if( !strcasecmp(optarg, "16:9") )
			a = 1;
			else if (!strcasecmp(optarg, "4:3") )
			a = 0;
			else
				{
				fprintf(stderr, "ERR:  Unrecognized aspect ratio: %s\n", optarg);
				return 1;
				}
			if( aspectflag != -1 )
				{
				fprintf(stderr,"ERR:  Already defined aspect ratio.\n");
				return 1;
				}

			aspectflag = a;
			break;
			}

		case 'A':
			{
			int a = -1;
			if( !strcasecmp(optarg, "ac3") )
			a = 0;
			else if (!strcasecmp(optarg, "mp2"))
			a = 2;
			else if( !strcasecmp(optarg, "pcm"))
			a = 4;
			else
				{
				fprintf(stderr, "ERR:  Cannot determine audio format: %s\n", optarg);
				return 1;
				}
			if( audioflag != -1 )
				{
				fprintf(stderr, "ERR:  Already defined audio format.\n");
				return 1;
				}
			audioflag = a;
			break;
			}

		case 'v':
			if( 1 != sscanf(optarg, "%d", &vtsnum) )
				{
				fprintf(stderr, "ERR:  Could not parse VTS number: %s\n", optarg);
				return 1;
				}
			break;
		case 'n':
			{
			int i;

			for(i = 1; i <= 99; i++ )
				{
				FILE *h;
				sprintf(realfbase, "%s_%02d_0.IFO", fbase, i);
				h = fopen(realfbase, "r");
				if( !h ) break;
				fclose(h);
				}
			fprintf(stderr, "STAT: Picking VTS %02d\n", i);
			vtsnum = i;
			break;
			}
		}
	}
if(optind == argc )
	{
	fprintf(stderr,\
	"syntax:\n\
	ifogen -l audio_language_codes -s subpicture_language_codes [use none for none] -o VTSBASE VOBFILE(s)\n");
	return 1;
	}

sprintf(realfbase, "%s_%02d", fbase, vtsnum);
FindVobus(realfbase, argc - optind, argv + optind, &va);
setattr();
WriteIFOs(realfbase, &va);
FixVobus(realfbase, &va);

return 0;
} /* end function main */

