/*
 * Copyright (C) 2002 Jan Panteltje panteltje@yahoo.com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <getopt.h>

#include "dvdauthor.h"


char *strsave(char *s) /*save char array s somewhere*/
{
char *p, *malloc();

p = malloc( strlen(s) +  1);
if(!p)
	{
	fprintf(stderr, "strsave(): could not allocate space for string %s\n", s);
	exit(1);
	}

strcpy(p, s);

return p;
} /* end function strsave */


int set_audio_mode(char *buffer, int pos, char *language_str)
{
int i;
int apos;
int audioflag;
int offset;
int audio_type;
char *token, *running;
char language_codes[100][2]; // 2 char per languages */

if(debug_flag)
	{
	fprintf(stderr, "set_audio_mode(): buffer=%p pos=%d\n", buffer, pos);
	}

/*
parse the language_str to get the names of the languages, and their number.
The format must be like this:
makedvd -l en,es,fr,it,de,pt,gr,he
for English, Spanish, French, Italien, German, Portuguese, Greek, Hebrew 
The characters are directly used.
*/ 


running = strsave(language_str);
if(! running)
	{
	exit(1);
    }

audio_streams = 0;
while(1)
	{
	token = strsep (&running, ",");
	if(token == NULL) break;

	/* avoid empty string */
	if(token[0] == 0) continue;
	fprintf(stderr, "set_audio_mode(): adding %c%c\n", token[0], token[1]);

	language_codes[audio_streams][0] = token[0];	
	language_codes[audio_streams][1] = token[1];	

	audio_streams++;
	} /* end while parse language_str */

fprintf(stderr, "set_audio_mode(): audio_streams=%d\n", audio_streams);

/*
byte 0:
bit 7 6 5 	coding mode 0=ac3 1=? 2=mpeg1 3=mpeg2ext 4=lpcm 5=? 6=dts 7=?
bit 4		multichannel ext present
bit 3 2		language type 0=unspecided 1=per bytes 2 and 3
bit 1 0		application mode 0=unspecified 1=karaoke 2=surround
*/
audioflag = (2 << 5) | (0 << 4) | (1 << 3) | 0; // mp2,  !multichext,  per byte 2&3

/*
byte 1:
bit 7 6		quantization, 0=16bpps 1=20bps 2=24bps 3=DRC (dynamic range control).
bit 5 4		sample rate 0=48kbps
bit 3		?
bit 2 1		channels - 1	 
*/
audio_type = (1 << 6) | (0 << 4); // 20bps, 48kbps 

if(pos == 0)
	{
	/* number of audio streams in VMGM_VOBS	number of audio streams in VTSM_VOBS */
	offset = 0x100;
	audio_streams = 0;
	}
else
	{
	/* number of audio streams in VTS_VOBS */
	offset = 0x200;
	}

/* number of audio streams in VTS_VOBS (2 bytes) */
buffer[offset + 2] = 0;
buffer[offset + 3] = audio_streams;

/* maximum 8 * 8 audio attributes VTS_VOBS */
for(i = 0; i < audio_streams; i++)
	{
	apos = offset + 4 + (i * 8);

	buffer[apos    ] = audioflag;				// bits 7,6,5 hold audio type
	buffer[apos + 1] = audio_type | 0; 			// channels -1 (for example mono is 0, stereo is 1)

	buffer[apos + 2] = language_codes[i][0];	// first language character
	buffer[apos + 3] = language_codes[i][1];	// second language character

	buffer[apos + 4] = 0;						// reserved language code extention
	buffer[apos + 5] = 0;						// code extention 0=unspecifed, 1=normal

	buffer[apos + 6] = 0;						// karaoke mode
	buffer[apos + 7] = 0;						// surround mode
	} /* end for all audio streams */

return 1;
} /* end function set_audio_mode */



int set_subpicture_mode(char *buffer, int pos, char *language_str)
{
int i;
int apos;
int offset;
char *token, *running;
char language_codes[100][2]; // 2 char per languages */

if(debug_flag)
	{
	fprintf(stderr, "set_subpicture_mode(): buffer=%p pos=%d\n", buffer, pos);
	}

/*
parse the language_str to get the names of the languages, and their number.
The format must be like this:
makedvd -l en,es,fr,it,de,pt,gr,he
for English, Spanish, French, Italien, German, Portuguese, Greek, Hebrew 
The characters are directly used.
*/ 

subpicture_streams = 0;
if(! language_str)
	{
	fprintf(stderr, "No subpicture streams.\n");

	return 1; /* no subs */
	}

running = strsave(language_str);
if(! running)
	{
	exit(1);
    }

while(1)
	{
	token = strsep (&running, ",");
	if(token == NULL) break;

	/* avoid empty string */
	if(token[0] == 0) continue;
	fprintf(stderr, "set_subpicture_mode(): adding %c%c\n", token[0], token[1]);

	language_codes[subpicture_streams][0] = token[0];	
	language_codes[subpicture_streams][1] = token[1];	


	subpicture_streams++;
	} /* end while parse language_str */

fprintf(stderr, "set_subpicture_mode(): subpicture_streams=%d\n", subpicture_streams);

/*
byte 0:
bit 7 6 5 	coding mode 0=2 bit rle.
bit 4 3 2	not used.	
bit 1 0		language type 0=unspecided 1=per bytes 2 and 3.

byte 1:     not used

byte 2:     1st character of language code.
byte 3:     2nd character of language code.
byte 4:     reserved for languuage code extention.
byte 5:     code extention ?? (SPRM #19)
*/

if(pos == 0)
	{
	/*
	number of subpicture streams in VMGM_VOBS (0 or 1), number of subpicture streams in VTSM_VOBS (0 or 1)
	*/
	offset = 0x100;
	subpicture_streams = 0;
	}
else
	{
	/* number of subpicture streams in VTS_VOBS */
	offset = 0x200;
	}

fprintf(stderr, "set_subpicture_mode(): offset=0x%x using subpicture_streams=%d\n",\
offset, subpicture_streams);

/* number of subpicture streams in VTS_VOBS (2 bytes) */
buffer[offset + 0x54] = 0;
buffer[offset + 0x55] = subpicture_streams;

/* maximum 32 * 6 subpicture attributes VTS_VOBS */
for(i = 0; i < subpicture_streams; i++)
	{
	apos = offset + 0x56 + (i * 6);

	buffer[apos    ] = 1;						// bit 7 6 5 rle coding, language type per bytes 2 and 3
	buffer[apos + 1] = 0;						// not used

	buffer[apos + 2] = language_codes[i][0];	// first language character
	buffer[apos + 3] = language_codes[i][1];	// second language character

//fprintf(stderr, "setting lc buffer[apos + 2-3] =%c%c\n",  buffer[apos + 2],  buffer[apos + 3]);

	buffer[apos + 4] = 0;						// reserved language code extention
	buffer[apos + 5] = 0;						// code extention
	} /* end for all subpicture streams */

return 1;
} /* end function set_subpicture_mode */



int set_video_mode()
{
int byte0, byte1;
int mpeg2;
int standard;
int aspect;
int disallow_auto_pan_scan;
int disallow_auto_letterbox;
int constant_bit_rate;
int resolution;
int letterboxed;
int film;

if(debug_flag)
	{
	fprintf(stderr, "set_video_mode(): arg none\n");
	}

/*
Byte 0:
 bit 7, 6 Coding mode:
  0 = mpeg-1, 1 = mpeg-2
 bit 5, 4 Standard:
  0 = NTSC, 1 = PAL
 bit 3, 2 Aspect
  0 = 4:3
  1 = not specified
  2 = reserved
  3 = 16:9
  1 and 2 should not be used
 bit 1 automatic Pan/Scan
  1 = disallowed
 bit 0 automatic Letterbox
   1 = disallowed
Byte 1:
 bit 7:
  CC for line 21 field 1 in GOP (NTSC only)
 bit 6:
  CC for line 21 field 2 in GOP (NTSC only)
 bit 5 not used
 bit 4 bit Rate
  0 = VBR 1 = CBR
 bit 3, 2 Resolution NTSC (PAL)
  0 = 720x480 (720x576)
  1 = 704x480 (704x576)
  2 = 352x480 (352x576)
  3 = 352x240 (352x288)
 bit 1 letterboxed
  1 = top and bottom cropped
 bit 0 Film
  0 = camera, 1 = film (PAL only)
*/

/* OK, we are mpeg2, PAL 720x576, vbr, 4:3 aspect. */
mpeg2 = 1;
standard = 1;					// PAL
aspect = 0;						// 4:3
disallow_auto_pan_scan = 0;
disallow_auto_letterbox = 1;
tvflag = 1; // GLOBAL 0 = ntsc, 1 = pal
constant_bit_rate = 0;			// ???? muxed as vbr!
resolution = 0;					// 720x576
letterboxed = 0;
film = 0;

byte0 = 0;
byte0 = \
	(mpeg2 << 6) | \
	(standard << 4) | \
	(aspect << 2) | \
	( disallow_auto_pan_scan << 1) | \
	disallow_auto_letterbox;

byte1 = 0;
byte1 = \
	(constant_bit_rate << 4) | \
	(resolution << 3) | \
	(letterboxed << 1) | \
	film;

video_mode0 = byte0;
video_mode1 = byte1;
videoattr = byte1 | (byte0 << 8);

fprintf(stderr,\
"set_video_mode(): setting byte0=%02x byte1=%02x videoattr=%04x\n",\
byte0, byte1, videoattr);

return 1;
} /* end function set_video_mode */

