/*
 * Copyright (C) 2002 Jan Panteltje <panteltje@yahoo.com>,
 * based on dvdauthor by Scott Smith (trckjunky@users.sourceforge.net) 
 * added multiple audio and subpicture channels.
 *
 * Modified and copy right Jan Panteltje 2002
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include <netinet/in.h>
#include <stdio.h>
#include <getopt.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

// #define SHOWDATA

int pos=0;
int queuedlen=0;

#define WAITLEN (32*1024)

#define BUFLEN (65536)

struct fdbuf {
    int pos,len;
    struct fdbuf *next;
    unsigned char buf[BUFLEN];
};

struct ofd {
    int fd;
    char *fname;
    struct fdbuf *firstbuf,**lastbufptr;
    int len;
} outputfds[256];

int ofdlist[256],numofd;

int closing=0;

fd_set rfd,wfd;

int dowork(int checkin)
{
    int i,n=-1;
    
    if( !numofd )
        return checkin;
    if( checkin ) {
        FD_SET(STDIN_FILENO,&rfd);
        n=STDIN_FILENO;
    } else {
        FD_CLR(STDIN_FILENO,&rfd);
    }
    while(1) {
        int minq=-1;
        for( i=0; i<numofd; i++ ) {
            struct ofd *o=&outputfds[ofdlist[i]];

            if( o->fd != -1 ) {
                if( o->fd == -2 ) {
                    int fd;
                    fd=open(o->fname,O_CREAT|O_WRONLY|O_NONBLOCK,0666);
                    if( fd == -1 && errno == ENXIO ) {
                        continue;
                    }
                    if( fd == -1 ) {
                        fprintf(stderr,"Cannot open %s: %s\n",o->fname,strerror(errno));
                        exit(1);
                    }
                    o->fd=fd;
                }
                // at this point, fd >= 0 
                if( minq == -1 || o->len < minq ) {
                    minq=o->len;
                }
                if( o->len > 0 ) {
                    if( o->fd>n )
                        n=o->fd;
                    FD_SET(o->fd,&wfd);
                } else {
                    FD_CLR(o->fd,&wfd);
                    if( closing ) {
                        close(o->fd);
                        o->fd=-1;
                    }
                }
            }
        }
        // if all the open files have more then WAITLEN bytes of data
        // queued up, then don't process anymore
        if( minq >= WAITLEN ) {
            FD_CLR(STDIN_FILENO,&rfd);
            break;
        } else if( minq >= 0 ) // as long as one file is open, continue
            break;
        sleep(1);
    }
    if( n == -1 )
        return 0;
    i=select(n+1,&rfd,&wfd,NULL,NULL);
    if( i > 0 ) {
        for( i=0; i<numofd; i++ ) {
            struct ofd *o=&outputfds[ofdlist[i]];
            if( o->fd >= 0 && FD_ISSET( o->fd, &wfd ) ) {
                struct fdbuf *f=o->firstbuf;
                n=write(o->fd,f->buf+f->pos,f->len-f->pos);
                if( n == -1 ) {
                    fprintf(stderr,"Error writing to fifo: %s\n",strerror(errno));
                    exit(1);
                }
                queuedlen-=n;
                f->pos+=n;
                if( f->pos == f->len ) {
                    o->firstbuf=f->next;
                    if( o->lastbufptr == &f->next )
                        o->lastbufptr=&o->firstbuf;
                    free(f);
                }
                o->len-=n;
            }
        }
        if( FD_ISSET( STDIN_FILENO, &rfd ) )
            return 1;
    }
    return 0;
}

int forceread(void *ptr,int len,FILE *h)
{
    while(!dowork(1));
    if( fread(ptr,1,len,h) != len ) {
        fprintf(stderr,"Could not read\n");
        closing=1;
        while( queuedlen )
            dowork(0);
        exit(1);
    }
    pos+=len;
    return len;
}

void writetostream(int stream,unsigned char *buf,int len)
{
    struct ofd *o=&outputfds[stream];

    if( o->fd == -1 )
        return;

    while( len > 0 ) {
        int thislen;
        struct fdbuf *fb;
        if( !o->lastbufptr[0] ) {
            o->lastbufptr[0]=malloc(sizeof(struct fdbuf));
            o->lastbufptr[0]->pos=0;
            o->lastbufptr[0]->len=0;
            o->lastbufptr[0]->next=0;
        }
        fb=o->lastbufptr[0];
        thislen=BUFLEN-fb->len;
        if( !thislen ) {
            o->lastbufptr=&fb->next;
            continue;
        }
        if( thislen > len )
            thislen=len;
        
        o->len+=thislen;
        memcpy(fb->buf+fb->len,buf,thislen);
        fb->len+=thislen;
        len-=thislen;
        buf+=thislen;
        queuedlen+=thislen;
    }
}

int main(int argc,char **argv)
{
    unsigned int hdr=0;
    unsigned char c,buf[200];
    int outputenglish=1,outputstream=0,oc;

    for( oc=0; oc<256; oc++ )
        outputfds[oc].fd=-1;

    while( -1 != (oc=getopt(argc,argv,"ha:v:o:")) ) {
        switch(oc) {
        case 'a':
        case 'v':
            if( outputstream ) {
                fprintf(stderr,"can only output one stream to stdout at a time\n; use -o to output more than\none stream\n");
                exit(1);
            }
            outputstream=((oc=='a')?0xc0:0xe0)+atoi(optarg);
            break;

        case 'o':
            if( !outputstream ) {
                fprintf(stderr,"no stream selected\n");
                exit(1);
            }
            outputfds[outputstream].fd=-2;
            outputfds[outputstream].fname=optarg;
            outputstream=0;
            break;

            // case 'h':
        default:
            fprintf(stderr,
                    "usage: mpeg2desc [options] < movie.mpg\n"
                    "\t-a #: output audio stream # to stdout\n"
                    "\t-v #: output video stream # to stdout\n"
                    "\t-h: help\n"
                );
            exit(1);
        }
    }
    if( outputstream ) {
        outputenglish=0;
        outputfds[outputstream].fd=STDOUT_FILENO;
    }
    numofd=0;
    for( oc=0; oc<256; oc++ )
        if( outputfds[oc].fd != -1 ) {
            ofdlist[numofd++]=oc;
            outputfds[oc].firstbuf=0;
            outputfds[oc].lastbufptr=&outputfds[oc].firstbuf;
            outputfds[oc].len=0;
        }
    FD_ZERO(&rfd);
    FD_ZERO(&wfd);    

    forceread(&hdr,4,stdin);
    while(1) {
        if( outputenglish )
            printf("%08x: ",pos-4);
        switch( ntohl(hdr) ) {
            // start codes:
        case 0x1b3: // sequence header
            if( outputenglish )
                printf("sequence hdr\n");
            forceread(buf,8,stdin);
            forceread(&hdr,4,stdin);
            break;

        case 0x1b5: // extension header
            if( outputenglish )
                printf("extension hdr\n");
            forceread(buf,1,stdin);
            switch( buf[0]>>4 ) {
            case 1:
                forceread(buf+1,5,stdin);
                break;
            case 2:
                forceread(buf+1,(buf[0]&1)?7:3,stdin);
                break;
            case 8:
                forceread(buf+1,4,stdin);
                if( buf[4]&64 )
                    forceread(buf+5,2,stdin);
                break;
            }
            forceread(&hdr,4,stdin);
            break;

        case 0x1b8: // group of pictures
            if( outputenglish )
                printf("GOP\n");
            forceread(buf,4,stdin);
            forceread(&hdr,4,stdin);
            break;

        case 0x1ba: // mpeg_pack_header
        {
            long scr,scrhi,scrext;
            long long fulltime;
            forceread(buf,10,stdin);
            if( buf[9]&7 ) {
                printf("WARN: pack padding\n");
            }
            if( (buf[0] & 0xC0) != 0x40 && outputenglish) {
                printf("WARN: not an mpeg2 pack header\n");
            }
            scrhi=(buf[0]&0x20)>>5;
            scr=((buf[0]&0x18)<<27)|
                ((buf[0]&3)<<28)|
                (buf[1]<<20)|
                ((buf[2]&0xf8)<<12)|
                ((buf[2]&3)<<13)|
                (buf[3]<<5)|
                ((buf[4]&0xf8)>>3);
            scrext=((buf[4]&3)<<7)|
                (buf[5]>>1);
            if( scrext >= 300 && outputenglish ) {
                printf("WARN: scrext in pack hdr > 300: %ld\n",scrext);
            }
            fulltime=((long long)scrhi)<<32|((long long)scr);
            fulltime*=300;
            fulltime+=scrext;
            if( outputenglish )
                printf("pack hdr, %lld.%02lld sec\n",fulltime/27000000,(fulltime%27000000)/(27000000/100));
            forceread(&hdr,4,stdin);
            break;
        }

        case 0x1bb: // mpeg_system_header
        case 0x1bd:
        case 0x1be:
        case 0x1bf:
        case 0x1c0:
        case 0x1c1:
        case 0x1c2:
        case 0x1c3:
        case 0x1c4:
        case 0x1c5:
        case 0x1c6:
        case 0x1c7:
        case 0x1c8:
        case 0x1c9:
        case 0x1ca:
        case 0x1cb:
        case 0x1cc:
        case 0x1cd:
        case 0x1ce:
        case 0x1cf:
        case 0x1d0:
        case 0x1d1:
        case 0x1d2:
        case 0x1d3:
        case 0x1d4:
        case 0x1d5:
        case 0x1d6:
        case 0x1d7:
        case 0x1d8:
        case 0x1d9:
        case 0x1da:
        case 0x1db:
        case 0x1dc:
        case 0x1dd:
        case 0x1de:
        case 0x1df:
        case 0x1e0:
        case 0x1e1:
        case 0x1e2:
        case 0x1e3:
        case 0x1e4:
        case 0x1e5:
        case 0x1e6:
        case 0x1e7:
        case 0x1e8:
        case 0x1e9:
        case 0x1ea:
        case 0x1eb:
        case 0x1ec:
        case 0x1ed:
        case 0x1ee:
        case 0x1ef:
        {
            unsigned char c;
            int ext=0,extra=0,readlen;

            c=ntohl(hdr);
            if( c == 0xBB ) {
                if( outputenglish )
                    printf("system header");
            } else if( c == 0xBD ) {
                if (outputenglish)
                    printf("pes private1");
                ext=1;
            } else if( c == 0xBE ) {
                if( outputenglish )
                    printf("pes padding");
            } else if( c == 0xBF ) {
                if( outputenglish )
                    printf("pes private2");
            } else if( c >= 0xC0 && c <= 0xDF ) {
                if( outputenglish )
                    printf("pes audio %d",c-0xC0);
                ext=1;
            } else if( c >= 0xE0 && c <= 0xEF ) {
                if( outputenglish )
                    printf("pes video %d",c-0xE0);
                ext=1;
            }
            forceread(buf,2,stdin); // pes packet length
            extra=(buf[0]<<8)|buf[1];
            if( outputenglish )
                printf("; length=%d",(buf[0]<<8)|buf[1]);
            readlen=forceread(buf,(extra>sizeof(buf))?sizeof(buf):extra,stdin);
            extra-=readlen;
            if( ext && (buf[1] & 128) && outputenglish ) {
                long long a1,a2,a3,pts;

                a1=(buf[3]&0xf)>>1;
                a2=((buf[4]<<8)|buf[5])>>1;
                a3=((buf[6]<<8)|buf[7])>>1;
                pts=(((long long)a1)<<30)|
                    (a2<<15)|
                    a3;
                printf("; pts %lld.%02lld sec",pts/90000,(pts%90000)/(90000/100));
            }
            if( ext && (buf[1] & 64) && outputenglish ) {
                long long a1,a2,a3,dts;
                int i;

                if( buf[1]&128 )
                    i=5;
                else
                    i=0;

                a1=(buf[i+3]&0xf)>>1;
                a2=((buf[i+4]<<8)|buf[i+5])>>1;
                a3=((buf[i+6]<<8)|buf[i+7])>>1;
                dts=(((long long)a1)<<30)|
                    (a2<<15)|
                    a3;
                printf("; dts %lld.%02lld sec",dts/90000,(dts%90000)/(90000/100));
            }
            if( outputenglish )
                printf("\n");
#ifdef SHOWDATA
            if( ext && outputenglish ) {
                int i=3+buf[2], j;
                printf("  ");
                for( j=0; j<16; j++ )
                    printf(" %02x",buf[j+i]);
                printf("\n");
            }
#endif
            if( ext ) {
                writetostream(ntohl(hdr)&255,buf+3+buf[2],readlen-3-buf[2]);
            }

            while( extra ) {
                readlen=forceread(buf,(extra>sizeof(buf))?sizeof(buf):extra,stdin);
                writetostream(ntohl(hdr)&255,buf,readlen);
                extra-=readlen;
            }
            forceread(&hdr,4,stdin);
            break;
        }

        default:
            if( outputenglish )
                printf("unknown hdr: %08x\n",ntohl(hdr));
            hdr>>=8;
            forceread(&c,1,stdin);
            hdr|=c<<24;
            break;
        }
    }
}
