/*
 * Copyright (C) 2002 Jan Panteltje <panteltje@yahoo.com>,
 * based on dvdauthor by Scott Smith (trckjunky@users.sourceforge.net) 
 * added multiple audio and subpicture channels.
 *
 * Modified and copy right Jan Panteltje 2002
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */



#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <dvdread/nav_read.h>
#include <dvdread/nav_print.h>


void FindVobus(const char *fname)
{
    static unsigned char buf[2048];
    FILE *h;
    int cursect=0;

    h=fopen(fname,"r");
    if( !h )
        return;
    while(1) {
        if( 2048 != fread(buf,1,2048,h) )
            break;
        if( buf[14] == 0 &&
            buf[15] == 0 &&
            buf[16] == 1 &&
            buf[17] == 0xbb && // system header
            buf[38] == 0 &&
            buf[39] == 0 &&
            buf[40] == 1 &&
            buf[41] == 0xbf && // 1st private2
            buf[1024] == 0 &&
            buf[1025] == 0 &&
            buf[1026] == 1 &&
            buf[1027] == 0xbf ) // 2nd private2
        {
            pci_t p;
	    dsi_t d;

            navRead_PCI(&p,buf+0x2d);
	    navRead_DSI(&d,buf+0x407);
            printf("\n\nVOBU: %x (%02x%02x%02x%02x%02x%02x)\n",cursect,buf[4]&255,buf[5]&255,buf[6]&255,buf[7]&255,buf[8]&255,buf[9]&255);
            navPrint_PCI(&p);
	    navPrint_DSI(&d);
        }
        cursect++;
    }
    fclose(h);
}

int main(int argc, char *argv[])
{
    if( argc != 2 ) {
        fprintf(stderr,"must pass vob as argument\n");
        return 1;
    }
    FindVobus(argv[1]);
    return 0;
}
